@@ -9,25 +9,31 @@ import matplotlib.pyplot as plt
 # generic class only implement algo. didn't implement detail func about M0, UT, F1 or F2
 class MfgBase(object):
     # it's a base class which implements the policy iteration and fixed point iteration
-    def __init__(
-        self, Q_guess=None, nv=0.05, Nx=200, Nt=200, T=1.0, alpha=1.5, zeta=0.5
-    ):
+    def __init__(self, Q_guess=None, nv=0.05, Nx=200, Nt=200, T=1.0, bc="periodic"):
         self.Nx = Nx
         self.Nt = Nt
-        self.xs = np.linspace(0, 1, self.Nx + 1)[:-1]  # grid on space
+        if bc == "periodic":
+            self.xs = np.linspace(0, 1, Nx + 1)[:-1]  # grid on space
+        elif bc == "dirichlet":
+            self.xs = np.linspace(0, 1, Nx + 1)[1:-1]  # grid on space
+        elif bc == "neumann":
+            self.xs = np.linspace(0, 1, Nx + 1)  # grid on space
+        self.Nx = self.xs.shape[0]
         self.dx = self.xs[1] - self.xs[0]
         self.ts = np.linspace(0, T, Nt + 1)  # grid on time
+        self.T = self.ts[-1]
         self.dt = self.ts[1] - self.ts[0]
         self.nv = nv  # diffusion paramter
         # H(Du,m) = 0.5*|Du|^2/F1(m) - F2(m)
-        #         = 0.5*|Du|^2/(1+m)^alpha - zeta*m
-        self.alpha = alpha
-        self.zeta = zeta
 
         # initail condition and terminal condition
         M0 = self.M0_func(self.xs)
         self.M0 = M0 / np.sum(M0) / self.dx
-        self.UT = self.UT_func(self.xs)
+        try:
+            self.UT = self.UT_func(self.xs)
+        except NotImplementedError:
+            # no use self.UT
+            pass
 
         # array to store solution
         self.U = np.zeros((self.Nt, self.Nx))  # t = 0 ... (T-dt)
@@ -51,35 +57,86 @@ class MfgBase(object):
         self.residual_HJB_history = []  # list to store history of residual
         self.sol_M_history = []  # list to store history of L-inf disatance
         self.sol_U_history = []  # list to store history of L-2 disatance
-        self.Laplace = self.build_Laplace()
-        self.DL, self.DR = self.build_D()
+        self.Laplace = self.build_Laplace(bc)
+        self.DL, self.DR = self.build_D(bc)
         self.A_fixp = sparse.eye(self.Nx) - self.dt * self.Laplace
         return
 
-    def build_Laplace(self):
-        dim_one_laplace = diags(
-            [1, 1, -2, 1, 1], [1 - self.Nx, -1, 0, 1, self.Nx - 1], format="csr"
-        )
-        dim_one_laplace = self.nv / self.dx**2 * dim_one_laplace
-        return dim_one_laplace
-
-    def build_D(self):
-        DL = (
-            diags(
-                [-1, 1, -1],
-                [-1, 0, self.Nx - 1],
+    def build_Laplace(self, bc="periodic"):
+        if bc == "periodic":
+            dim_one_laplace = diags(
+                [1, 1, -2, 1, 1],
+                [1 - self.Nx, -1, 0, 1, self.Nx - 1],
                 shape=(self.Nx, self.Nx),
                 format="csr",
             )
-            / self.dx
-        )
-        DR = (
-            diags(
-                [1, -1, 1], [1 - self.Nx, 0, 1], shape=(self.Nx, self.Nx), format="csr"
+            dim_one_laplace = self.nv / self.dx**2 * dim_one_laplace
+            return dim_one_laplace
+        elif bc == "dirichlet":
+            dim_one_laplace = diags(
+                [1, -2, 1], [-1, 0, 1], shape=(self.Nx, self.Nx), format="csr"
             )
-            / self.dx
-        )
-        return DL, DR
+            dim_one_laplace = self.nv / self.dx**2 * dim_one_laplace
+            return dim_one_laplace
+        elif bc == "neumann":
+            dim_one_laplace = diags(
+                [1, -2, 1], [-1, 0, 1], shape=(self.Nx, self.Nx), format="csr"
+            )
+            dim_one_laplace[0, 1] += 1
+            dim_one_laplace[-1, -2] += 1
+            dim_one_laplace = self.nv / self.dx**2 * dim_one_laplace
+            return dim_one_laplace
+        else:
+            raise NotImplementedError("unsupported boundary condition")
+
+    def build_D(self, bc="periodic"):
+        if bc == "periodic":
+            DL = (
+                diags(
+                    [-1, 1, -1],
+                    [-1, 0, self.Nx - 1],
+                    shape=(self.Nx, self.Nx),
+                    format="csr",
+                )
+                / self.dx
+            )
+            DR = (
+                diags(
+                    [1, -1, 1],
+                    [1 - self.Nx, 0, 1],
+                    shape=(self.Nx, self.Nx),
+                    format="csr",
+                )
+                / self.dx
+            )
+            return DL, DR
+        elif bc == "dirichlet":
+            DL = (
+                diags(
+                    [-1, 1],
+                    [-1, 0],
+                    shape=(self.Nx, self.Nx),
+                    format="csr",
+                )
+                / self.dx
+            )
+            DR = (
+                diags([-1, 1], [0, 1], shape=(self.Nx, self.Nx), format="csr") / self.dx
+            )
+            return DL, DR
+        elif bc == "neumann":
+            DL = (
+                diags([-1, 1], [-1, 0], shape=(self.Nx, self.Nx), format="csr")
+                / self.dx
+            )
+            DR = (
+                diags([-1, 1], [0, 1], shape=(self.Nx, self.Nx), format="csr") / self.dx
+            )
+            DL[0, 1] += -1 / self.dx
+            DL[-1, -2] += -1 / self.dx
+            return DL, DR
+        else:
+            raise NotImplementedError("unsupported boundary condition")
 
     # initial condition
     def M0_func(self, x):
@@ -89,6 +146,13 @@ class MfgBase(object):
     def UT_func(self, x):
         raise NotImplementedError(self.__class__.__name__ + "didn't implement UT_func")
 
+    def get_UT(self, mT):
+        # a defualt function to get UT
+        # if you want to use a different UT, you can rewrite this function
+        # especially, if UT is a function of x and M at end time
+        # like UT = np.sum(self.xs*mT)*self.dx*self.xs
+        return self.UT
+
     # hamilton
     # H(Du,m) = 0.5*|Du|^2/F1(m) - F2(m)
     def F1_func(self, m):
@@ -157,7 +221,8 @@ class MfgBase(object):
         this method should only update self.U
         """
         QL, QR = Q  # unpack
-        un = self.UT
+        mT = M[-1, :]
+        un = self.get_UT(mT)
         for ti in range(self.Nt - 1, -1, -1):
             mn = M[ti, :]
             ql = QL[ti, :]
@@ -177,7 +242,8 @@ class MfgBase(object):
 
         this method should only update self.U
         """
-        un = self.UT
+        mT = M[-1, :]
+        un = self.get_UT(mT)
         for ti in range(self.Nt - 1, -1, -1):
             u_old = self.U_old[ti, :]
             dul_old = np.maximum(self.DL @ u_old, 0.0)
@@ -373,7 +439,8 @@ class MfgBase(object):
 
     def residual_HJB(self, M, U):
         residual_ = np.zeros_like(U)  # residual in every point
-        un = self.UT
+        mT = M[-1, :]
+        un = self.get_UT(mT)
         for ti in range(self.Nt - 1, -1, -1):
             mn = M[ti, :]
             u = U[ti, :]
@@ -390,39 +457,58 @@ class MfgBase(object):
 
     def compute_cost(self):
         cost = 0.0
+        mT = self.M[-1, :]
+        uT = self.get_UT(mT)
         for ti in range(self.Nt):
             m = self.M[ti, :]
             u = self.U[ti, :]
             ql = np.maximum((self.DL @ u) / self.F1_func(m), 0.0)
             qr = np.minimum((self.DR @ u) / self.F1_func(m), 0.0)
             q2 = ql**2 + qr**2
-            cost += np.sum(self.dt * self.dx * m * (0.5 * self.F1_func(m) * q2 + self.F2_func(m)))
-        cost += np.sum(self.dx * self.UT * self.M[-1, :])
+            cost += np.sum(
+                self.dt * self.dx * m * (0.5 * self.F1_func(m) * q2 + self.F2_func(m))
+            )
+        cost += np.sum(self.dx * uT * self.M[-1, :])
         return cost
 
 
 class MfgBase2D(object):
     # it's a base class which implements the policy iteration and fixed point iteration
     def __init__(
-        self, Q_guess=None, nv=0.05, Nx=200, Nt=200, T=1.0, alpha=0.0, zeta=0.0
+        self,
+        Q_guess=None,
+        nv=0.05,
+        Nx=200,
+        Nt=200,
+        T=1.0,
+        bc="periodic",
     ):
-        self.Nx = Nx
         self.Nt = Nt
-        self.xs1 = np.linspace(0, 1, self.Nx + 1)[:-1]  # grid on space
-        self.xs2 = np.linspace(0, 1, self.Nx + 1)[:-1]  # grid on space
+        if bc == "periodic":
+            self.xs1 = np.linspace(0, 1, Nx + 1)[:-1]  # grid on space
+            self.xs2 = np.linspace(0, 1, Nx + 1)[:-1]  # grid on space
+        elif bc == "dirichlet":
+            self.xs1 = np.linspace(0, 1, Nx + 1)[1:-1]  # grid on space
+            self.xs2 = np.linspace(0, 1, Nx + 1)[1:-1]  # grid on space
+        elif bc == "neumann":
+            self.xs1 = np.linspace(0, 1, Nx + 1)  # grid on space
+            self.xs2 = np.linspace(0, 1, Nx + 1)  # grid on space
+        self.Nx = self.xs1.shape[0]
         self.dx = self.xs1[1] - self.xs1[0]
         self.ts = np.linspace(0, T, Nt + 1)  # grid on time
         self.dt = self.ts[1] - self.ts[0]
+        self.T = self.ts[-1]
         self.nv = nv  # diffusion paramter
         # H(Du,m) = 0.5*|Du|^2/F1(m) - F2(m)
-        #         = 0.5*|Du|^2/(1+m)^alpha - zeta*m
-        self.alpha = alpha
-        self.zeta = zeta
 
         # initail condition and terminal condition
         M0 = self.M0_func(self.xs1, self.xs2)
         self.M0 = M0 / np.sum(M0) / (self.dx) ** 2
-        self.UT = self.UT_func(self.xs1, self.xs2)
+        try:
+            self.UT = self.UT_func(self.xs1, self.xs2)
+        except NotImplementedError:
+            # no use self.UT
+            pass
 
         # array to store solution
         self.U = np.zeros((self.Nt, (self.Nx) ** 2))  # t = 0 ... (T-dt)
@@ -454,33 +540,76 @@ class MfgBase2D(object):
         self.residual_HJB_history = []  # list to store history of residual
         self.sol_M_history = []  # list to store history of L-inf disatance
         self.sol_U_history = []  # list to store history of L-2 disatance
-        self.Laplace = self.build_Laplace()
-        self.DL1, self.DR1, self.DL2, self.DR2 = self.build_D()
+        self.Laplace = self.build_Laplace(bc)
+        self.DL1, self.DR1, self.DL2, self.DR2 = self.build_D(bc)
         self.A_fixp = sparse.eye(self.Nx**2) - self.dt * self.Laplace
         return
 
-    def build_Laplace(self):
-        dim_one_laplace = diags(
-            [1, 1, -2, 1, 1], [1 - self.Nx, -1, 0, 1, self.Nx - 1], format="csr"
-        )
-        dim_two_laplace = sparse.kron(
-            dim_one_laplace, sparse.eye(self.Nx)
-        ) + sparse.kron(sparse.eye(self.Nx), dim_one_laplace)
-        dim_two_laplace = self.nv / self.dx**2 * dim_two_laplace
-        return dim_two_laplace
+    def build_Laplace(self, bc="periodic"):
+        if bc == "periodic":
+            dim_one_laplace = diags(
+                [1, 1, -2, 1, 1], [1 - self.Nx, -1, 0, 1, self.Nx - 1], format="csr"
+            )
+            dim_two_laplace = sparse.kron(
+                dim_one_laplace, sparse.eye(self.Nx)
+            ) + sparse.kron(sparse.eye(self.Nx), dim_one_laplace)
+            dim_two_laplace = self.nv / self.dx**2 * dim_two_laplace
+            return dim_two_laplace
+        elif bc == "dirichlet":
+            dim_one_laplace = diags([1, -2, 1], [-1, 0, 1], format="csr")
+            dim_two_laplace = sparse.kron(
+                dim_one_laplace, sparse.eye(self.Nx)
+            ) + sparse.kron(sparse.eye(self.Nx), dim_one_laplace)
+            dim_two_laplace = self.nv / self.dx**2 * dim_two_laplace
+            return dim_two_laplace
+        elif bc == "neumann":
+            dim_one_laplace = diags([1, -2, 1], [-1, 0, 1], format="csr")
+            dim_one_laplace[0, 1] += 1
+            dim_one_laplace[-1, -2] += 1
+            dim_two_laplace = sparse.kron(
+                dim_one_laplace, sparse.eye(self.Nx)
+            ) + sparse.kron(sparse.eye(self.Nx), dim_one_laplace)
+            dim_two_laplace = self.nv / self.dx**2 * dim_two_laplace
+            return dim_two_laplace
+        else:
+            raise NotImplementedError("unsupported boundary condition")
 
-    def build_D(self):
-        DL = diags(
-            [-1, 1, -1], [-1, 0, self.Nx - 1], shape=(self.Nx, self.Nx), format="csr"
-        )
-        DR = diags(
-            [1, -1, 1], [1 - self.Nx, 0, 1], shape=(self.Nx, self.Nx), format="csr"
-        )
-        DL1 = sparse.kron(sparse.eye(self.Nx), DL) / self.dx
-        DR1 = sparse.kron(sparse.eye(self.Nx), DR) / self.dx
-        DL2 = sparse.kron(DL, sparse.eye(self.Nx)) / self.dx
-        DR2 = sparse.kron(DR, sparse.eye(self.Nx)) / self.dx
-        return DL1, DR1, DL2, DR2
+    def build_D(self, bc="periodic"):
+        if bc == "periodic":
+            DL = diags(
+                [-1, 1, -1],
+                [-1, 0, self.Nx - 1],
+                shape=(self.Nx, self.Nx),
+                format="csr",
+            )
+            DR = diags(
+                [1, -1, 1], [1 - self.Nx, 0, 1], shape=(self.Nx, self.Nx), format="csr"
+            )
+            DL1 = sparse.kron(sparse.eye(self.Nx), DL) / self.dx
+            DR1 = sparse.kron(sparse.eye(self.Nx), DR) / self.dx
+            DL2 = sparse.kron(DL, sparse.eye(self.Nx)) / self.dx
+            DR2 = sparse.kron(DR, sparse.eye(self.Nx)) / self.dx
+            return DL1, DR1, DL2, DR2
+        elif bc == "dirichlet":
+            DL = diags([-1, 1], [-1, 0], shape=(self.Nx, self.Nx), format="csr")
+            DR = diags([-1, 1], [0, 1], shape=(self.Nx, self.Nx), format="csr")
+            DL1 = sparse.kron(sparse.eye(self.Nx), DL) / self.dx
+            DR1 = sparse.kron(sparse.eye(self.Nx), DR) / self.dx
+            DL2 = sparse.kron(DL, sparse.eye(self.Nx)) / self.dx
+            DR2 = sparse.kron(DR, sparse.eye(self.Nx)) / self.dx
+            return DL1, DR1, DL2, DR2
+        elif bc == "neumann":
+            DL = diags([-1, 1], [-1, 0], shape=(self.Nx, self.Nx), format="csr")
+            DR = diags([-1, 1], [0, 1], shape=(self.Nx, self.Nx), format="csr")
+            DL[0, 1] += 1
+            DR[-1, -2] += 1
+            DL1 = sparse.kron(sparse.eye(self.Nx), DL) / self.dx
+            DR1 = sparse.kron(sparse.eye(self.Nx), DR) / self.dx
+            DL2 = sparse.kron(DL, sparse.eye(self.Nx)) / self.dx
+            DR2 = sparse.kron(DR, sparse.eye(self.Nx)) / self.dx
+            return DL1, DR1, DL2, DR2
+        else:
+            raise NotImplementedError("unsupported boundary condition")
 
     # initial condition
     # should return array with shape (self.Nx*self.Nx, )
@@ -492,6 +621,13 @@ class MfgBase2D(object):
     def UT_func(self, x1, x2):
         raise NotImplementedError(self.__class__.__name__ + "didn't implement UT_func")
 
+    def get_UT(self, mT):
+        # a defualt function to get UT
+        # if you want to use a different UT, you can rewrite this function
+        # especially, if UT is a function of x and M at end time
+        # like UT = np.sum(self.xs*mT)*self.dx**2*self.xs
+        return self.UT
+
     # hamilton
     # H(Du,m) = 0.5*|Du|^2/F1(m) - F2(m)
     def F1_func(self, m):
@@ -565,7 +701,8 @@ class MfgBase2D(object):
 
     def solve_HJB(self, M, Q):
         QL1, QR1, QL2, QR2 = Q  # unpack
-        un = self.UT
+        mT = M[-1, :]
+        un = self.get_UT(mT)
         for ti in range(self.Nt - 1, -1, -1):
             mn = M[ti, :]
             ql1 = QL1[ti, :]
@@ -587,7 +724,8 @@ class MfgBase2D(object):
 
         this method should only update self.U
         """
-        un = self.UT
+        mT = M[-1, :]
+        un = self.get_UT(mT)
         for ti in range(self.Nt - 1, -1, -1):
             u_old = self.U_old[ti, :]
             dul1_old = np.maximum(self.DL1 @ u_old, 0.0)
@@ -802,7 +940,8 @@ class MfgBase2D(object):
 
     def residual_HJB(self, M, U):
         residual_ = np.zeros_like(U)  # residual in every point
-        un = self.UT
+        mT = M[-1, :]
+        un = self.get_UT(mT)
         for ti in range(self.Nt - 1, -1, -1):
             mn = M[ti, :]
             u = U[ti, :]
@@ -820,6 +959,8 @@ class MfgBase2D(object):
 
     def compute_cost(self):
         cost = 0.0
+        mT = M[-1, :]
+        uT = self.get_UT(mT)
         for ti in range(self.Nt):
             m = self.M[ti, :]
             u = self.U[ti, :]
@@ -828,8 +969,13 @@ class MfgBase2D(object):
             ql2 = np.maximum((self.DL2 @ u) / self.F1_func(m), 0.0)
             qr2 = np.minimum((self.DR2 @ u) / self.F1_func(m), 0.0)
             q2 = ql1**2 + qr1**2 + ql2**2 + qr2**2
-            cost += np.sum(self.dt * self.dx**2 * m * (0.5 * self.F1_func(m) * q2 + self.F2_func(m)))
-        cost += np.sum(self.dx**2 * self.UT * self.M[-1, :])
+            cost += np.sum(
+                self.dt
+                * self.dx**2
+                * m
+                * (0.5 * self.F1_func(m) * q2 + self.F2_func(m))
+            )
+        cost += np.sum(self.dx**2 * uT * self.M[-1, :])
         return cost
 
 
